use log::info;
use quick_xml::{
    events::{attributes::Attributes, BytesDecl, BytesEnd, BytesStart, Event},
    Writer,
};
use rustls_pemfile::Item;
use tokio::io::AsyncWrite;

use crate::{
    error::StreamError,
    feature::Feature,
    tag::{self, Tag},
};

type Result<T> = std::result::Result<T, StreamError>;

pub async fn error<W: AsyncWrite + Unpin>(writer: W, err: StreamError) -> Result<()> {
    let mut writer = Writer::new(writer);
    let err = err.to_string();
    writer
        .write_event_async(Event::Start(BytesStart::new(tag::ERROR_ELEMENT)))
        .await?;
    writer
        .write_event_async(Event::Start(
            BytesStart::new(&err)
                .with_attributes(vec![("xmlns", "urn:ietf:params:xml:ns:xmpp-streams")]),
        ))
        .await?;
    writer
        .write_event_async(Event::End(BytesEnd::new(&err)))
        .await?;
    writer
        .write_event_async(Event::End(BytesEnd::new(tag::ERROR_ELEMENT)))
        .await?;
    Ok(())
}

pub async fn write_stream_header<W: AsyncWrite + Unpin>(writer: W, req: StreamAttrs) -> Result<()> {
    let mut writer = Writer::new(writer);
    writer
        .write_event_async(Event::Decl(BytesDecl::new("1.0", Some("utf-8"), None)))
        .await?;
    writer
        .write_event_async(Event::Start(
            BytesStart::new("stream:stream").with_attributes(vec![
                ("from", req.from.as_str()),
                ("to", req.to.as_str()),
                ("xmlns:stream", "http://etherx.jabber.org/streams"),
                ("xml:lang", "en"),
                ("version", "1.0"),
            ]),
        ))
        .await?;

    Ok(())
}

#[derive(Debug, Clone)]
pub struct StreamAttrs {
    pub from: String,
    pub to: String,
    pub namespace: XMLNamespace,
}

impl TryFrom<Attributes<'_>> for StreamAttrs {
    type Error = StreamError;

    fn try_from(value: Attributes<'_>) -> std::result::Result<Self, Self::Error> {
        let mut from: Option<String> = None;
        let mut to: Option<String> = None;
        let mut ns: Option<XMLNamespace> = None;
        for v in value {
            let v = v?;
            match v.key.local_name().into_inner() {
                b"from" => {
                    from = Some(String::from_utf8(v.value.to_vec())?);
                }
                b"to" => {
                    to = Some(String::from_utf8(v.value.to_vec())?);
                }
                b"xmlns" => match v.value.to_vec().as_slice() {
                    b"jabber:client" => {
                        ns = Some(XMLNamespace::JabberClient);
                    }
                    _ => return Err(StreamError::InvalidNamespace),
                },
                other => {
                    info!(
                        "ignoring key {}",
                        String::from_utf8(other.to_vec()).unwrap_or_default()
                    );
                }
            }
        }
        Ok(StreamAttrs {
            from: from.ok_or(StreamError::InvalidFrom)?,
            to: to.ok_or(StreamError::HostUnknown)?,
            namespace: ns.ok_or(StreamError::BadNamespacePrefix)?,
        })
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum XMLNamespace {
    JabberClient,
}
